/* ADH_list.cpp (C) 2005  adolfo@di-mare.com  */

/** \file  ADH_list.cpp
    \brief Implementaciones para la clase \c "ADH::list"

    \author Adolfo Di Mare <adolfo@di-mare.com>
    \date   2005
*/


#include "ADH_list.h"

namespace ADH { // namespace ADH {

/** Verifica la invariante de la clase \c "list".
    - Regresa \c "true" si \c "L" contiene un valor correcto.
    - Podra retornar \c "true" an cuando \c "L" tenga su valor corrupto.
    - Podra no retornar si \c "L" tiene su valor corrupto.

    \par Rep Diagrama de la clase
    \code
      L                                        NULL
    +---+    +-----+---+   +-----+---+   +-----+---+
    | *-|--->| v_1 | *-|-->| v_2 | *-|-->| v_3 | 0 |
    +---+    +-----+---+   +-----+---+   +-----+---+
   m_first    m_val next                      /\
                                              ||
                                         last_node(L)
    \endcode

    \par Descripcin en palabras
    - Una lista es una secuencia de nodos enlazados del primero al ltimo.
    - El ltimo nodo tiene su campo \c "next" en nulo [puntero \c (0)].
    - Para que el valor de la lista sea correcto es necesario que
      cada uno de los valores almacenados en la lista sea correcto.
    - El campo \c "next" no se llama \c "m_next" porque en los libros
      de texto el nombre que aparece siempre es \c "next".
    - El campo "next" debiera llamarse "m_next", como lo manda la convencin para nombrar
      los campos de una clase.
    - En este caso, se ha mantenido el nombre "next" sin el prefijo "m_" porque en la
      mayor parte de los libros de texto, el nombre de este campo es "next", de manera
      que tiene un valor mnemnico-pedaggico usar "next" en lugar de "m_next".

    \par Descripcin en smbolos
    - (this->m_first == 0) || (this->m_first == p ) apunta a un nodo
    - Para cada node de una lista
      (p->m_val es correcto && (p->next == 0 || p->next apunta a un node vlido )

    \pre Obliga al programador cliente a implementar la funcin
          \c bool check_ok( const value_type & );
*/
bool check_ok( const list & L ) {
    if ( ! (&L != 0) ) {
        /// - Invariante: ningn objeto puede estar almacenado en la posicin nula.
        return false;
    }

    if ( L.m_first == 0 ) {
        /// - Invariante: la lista vaca siempre se representa con el puntero nulo en \c m_first.
        return true;
    }

    list::node * p = L.m_first;
    while ( p != 0 ) {
        if ( ! check_ok( p->m_val ) ) {
            /// - Invariante: El valor \c "m_val" almacenado en un nodo de la lista
            ///   siempre debe ser un valor correcto.
            return false;
        }
        p = p->next;
    }
    return true;
}

/// Agrega una copia del valor \c "v" al principio de la lista \c "*this".
void list::push_front( const list::value_type& v ) {
    if (m_first == 0) {
        this->m_first = new node(v);
        m_first->next = 0;  // (*m_first).next = 0;
    }
    else {
        node *p;
        p = new node(v);
        p->next = m_first;  // (*p).next = (*this).m_first;
        m_first = p; // this->m_first = p;
    }
} // list::push_front()

/** Agrega una copia del valor \c "v" al final de la lista \c "*this".
    \par Complejidad
    - O ( \c size() )
*/
void list::push_back( const value_type& v ) {
    if ( empty() ) { // if (m_first == 0) ... // exactamente equivalente
        m_first = new node(v);
        m_first->next = 0;  // (*m_first).next = 0;
    }
    else {
        node * p = last_node( );
        p->next = new node(v);
        p->next->next = 0;
    }
} // list::push_back()

/** Elimina el valor del principio de la lista.
    \pre
    \c !empty()
*/
void list::pop_front() {
    #if defined(NDEBUG)
    //  #define NDEBUG ==> Activo para obtener al versin de produccin
        if ( empty() ) {
            throw (BOOOM!!!);
        }
    #endif

    node *p;
    p = m_first;    // recuerda adnde est el siguiente nodo
    m_first = m_first->next;
    delete p;      // elimina el nodo que fue cabeza de la lista
} //  list::pop_front()

/** Elimina el valor del final de la lista.
    \pre
    \c !empty()

    \par Complejidad
    - O ( \c size() )
*/
void list::pop_back() {
    #if defined(NDEBUG)
        if ( empty() ) {
            throw (BOOOM!!!);
        }
    #endif

    // avanza hasta el final   O---O---O---O
    node *q = 0;       //          |   |
    node *p = m_first; //          q   p    // "q" es el valor anterior de "p"
    while (p->next != 0) {
        q = p;
        p = p->next;
    }

    if (q == 0) { // nunca se entr al ciclo while()
        // Slo queda un node en la lista
        delete m_first; // mata al node
        m_first = 0;    // marca la lista como vaca
    }
    else {
       q->next = 0;
       delete p;
    }
} //  list::pop_back()

list::size_type list::size( ) const {
    size_type n = 0;
    list::iterator p = begin();
    while (p != end()) {
        n++;
        p++;
    }
    return n;
}

void list::clear() {
    #if 0
        while ( ! empty() ) { // No se le meta al Rep
            pop_front();      // - es "menos" eficiente
        }
    #endif

    #if 1
        while ( ! empty() ) {  // SI se le meta al Rep
            node *p = m_first;     // a la larga vale la pena que
            m_first = m_first->next;  // pop_front() sea inline...
            delete p;
        }
    #endif

    #ifdef ESTO_NUNCA_FUNCIONA
        while ( ! empty() ) {  // SI se le meta al Rep
            delete m_first;
            m_first = m_first->next;  // m_first->next puede estar "borrado"
        }
    #endif
} // list::clear()

/** Retorna el puntero al ltimo nodo de la lista.
    \pre
    <code>!L.empty()</code>
*/
list::node * list::last_node( ) const {
    list::node *p = m_first;
    while (p->next != 0) {
        p = p->next;
    }
    return p;
} // last_node()

/// Retorna un iterador que denota al ltimo valor de a lista \c "L".
/// - Retorna \c L.end() si la lista est vaca.
list::iterator last( const list & L ) {
    list::iterator p,q;
    q = L.end();
    p = L.begin();
    while ( p != L.end() ) {
        q = p; // "q" est "antes" de "p"
        ++p;
    }
    return q;
}

/** Retorna el puntero al nodo anterior.
    - Devuelve un puntero al nodo anterior al que \c "p" apunta.
    - Retorna \c "0" si \c "p" es el primer nodo de la lista.

    \pre
    - \c "*p" debe ser uno de los nodos de la lista
    - <code>(p != 0) && (! empty())</code>

    \remark
    Como la lista est enlazada en una sla direccin, para
    encontrar el antecesor  de un nodo es necesario recorrer
    toda la lista. Si se necesita usar una lista para la que
    el tiempo de ejecucin de esta operacin sea O(1), es
    necesario usar una lista doblemente enlazada.
*/
list::node * list::prev_node( const list::node * p ) const {
    if (p == m_first) {
        return 0;
    }

    list::node* q;
    q  = m_first; // busca desde el principio de la lista
    while (q != 0) {
        if ( p == q->next ) {
            return q;
        }
        q = q->next;
    }

    return 0;
/*  NOTA: en la especificacin de este mtodo se habla de
    "nodos", pese a que los nodos son parte del Rep de la
    lista. La razn por la que esto no es incorrecto es
    que este mtodo es un mtodo privado, usado en la
    implementacin. A pesar de este hecho, la definicin
    de qu hace este mtodo est separada de cmo lo
    hace por lo que esta especificacin permite varias
    implementaciones diferentes.
*/
} // prev_node()

list& list::operator=(const list& LO) {
    if (&LO == this) {
        return *this; // evita autocopia
    }

    // borra todo
    clear();

    // copia
    node *p = LO.m_first;
    while (p != 0) {
        push_back( p->m_val );
        p = p->next;
    }
    return *this;
/*  NOTA
    Esta implementacin es un poco lerda, pues sera ms eficiente
    no devolver todos los nodos de *this, y reutilizarlos para
    almacenar ah la copia de los valores que estn en "LO".

    Tambin sera ms eficiente no utilizar el push_back() para
    agregar los valores al final de "*this", lo que requerira
    mantener un puntero "m_last" al ltimo nodo agregado.

    - Esta lista es ms didctica que eficiente...
*/
} // list::operator=()

/// Intercambia <code>*this <==> L</code>
void list::swap(list& L) {
    node *tmp = L.m_first;
    L.m_first = m_first;
    m_first = tmp;
}

/// Traslada el valor de \c "LO" a \c "*this"
void list::move(list& LO) {
    if (&LO == this) {
        return; // evita autocopia
    }

    // borra todo
    clear();

    m_first    = LO.m_first;    // le roba a LO todos sus valores
    LO.m_first = 0;
}

#include <assert.h>

/** Agrega el valor \c "v" al contenedor.
    - El valor agregado queda antes de la posicin \c "p" en el contenedor.
    - Si <code>p==this->end()</code> el valor queda al final de la lista.

    \par Complejidad
    - O ( \c size() )
*/
void list::insert(list::iterator p, const value_type& v) {
    node* q = new node(v);
    if ( m_first==0 ) { // push_front(v)
        q->next = 0;
        m_first = q;
    }
    else if ( m_first==p.m_p ) { // push_front(v)
        q->next = m_first;
        m_first = q;
    }
    else if ( p.m_p==0 ) { // push_back(v)
        node *last = last_node( );
        q->next    = 0;
        last->next = q;
    }
    else {
        node *prev = prev_node(p.m_p);
        q->next = prev->next;
        prev->next = q;
    }
} // list::insert()

#include <assert.h>

/// Elimina del contenedor cada uno de los valores iguales a \c "v"
void list::remove(const value_type& v) {
    if( empty() ) {
        return; // no hay nada que borrar si la lista est vaca
    }

    assert( m_first != 0 ); // la lista contiene al menos un nodo

    while ( m_first->m_val == v ) { // borra a todos los "v" del principio
        node * p = m_first;
        m_first  = m_first -> next;
        delete p; // mata "p"
        if (m_first == 0) {
            return; // se sale si se le acaba la lista
        }
    }

    assert( m_first != 0 );        // la lista contiene al menos un nodo
    assert( m_first->m_val != v ); // ese nodo NO contiene el valor "v"

    // avanza hasta el final   O---O---O---O
    node *q;           //          |   |
    node *p = m_first; //          q   p    // "q" es el valor anterior de "p"
    while (p->next != 0) {
        q = p;
        p = p->next;
        if (p->m_val == v) {
            q->next = p->next;  // se brinca "p" desde "q"
            delete p;           // mata "p"
            p = q;
        }
    }
    assert( ! empty() );
} // list::remove()

#include <assert.h>

/** Traslada el valor \c "from" de la lista \c "L" y lo coloca antes de \c "p".
    - El valor \c "from" deja de ser parte de \c "L" y pasa a estar almacenado
      dentro de \c "*this".
    - Todos los iteradores que denotan \c "from" quedan con su valor
      invalidado.

    \pre
    - El valor \c "p" debe denotar una posicin en \c "*this" o \c "end()".
    - El valor \c "from" debe denotar una posicin en \c "L".

    \remark
    - Al trasladar el valor se evita copiarlo.

    \par Complejidad
    - O ( \c size() )
*/
void list::splice( iterator p, list& L, iterator from ) {
    assert( !false && "list::splice() ==> NO IMPLEMENTADO" );

    // Caso especial: [L == (1)] <==> L.splice(L.begin(), L, L.begin())
    if ( p.m_p == from.m_p ) {
        return;
    }
    // desconecta "from" de "L"
    node* q = L.prev_node(from.m_p);
    if ( q == 0 ) {
        L.m_first = from.m_p->next;
    }
    else {
        q->next = from.m_p->next;
    }

    if (p.m_p == 0) {          // pone "from" al final de "*this"
        q = last_node( );
        q->next = from.m_p;
        from.m_p->next = 0;
    }
    else if (p.m_p == m_first) { // pone "from" al principio de "*this"
        from.m_p->next = m_first;
        m_first = from.m_p;
    }
    else {
        q = prev_node( p.m_p ); // pone "from" antes de "p"
        from.m_p->next = p.m_p;
        q->next = from.m_p;
    }
} // list::splice()

/// <code>L == LL</code>?
bool operator == (const list& L, const list& LL) {
    if ( L.empty() && LL.empty() ) {
        return true;
    }

    if ( L.empty() || LL.empty() ) {
        return false;
    }

    list::iterator p =  L.begin();
    list::iterator q = LL.begin();
    do {
        if ( *p != *q ) {
            return false;
        }
        ++p;
        ++q;
    } while ( p != L.end() && q != LL.end() );

    return  ( p == L.end() && q == LL.end() );
}

/** Graba el valor de \c "L" en el flujo \c "COUT".
    - En particular, este es el operador que se invoca
      cuando se usa, por ejemplo, este tipo de instruccin:
      \code
      COUT << L << LO;
      \endcode
*/
std::ostream& operator<< (std::ostream &COUT, const list& L) {
#if 1
    COUT << "(";
    list::node *p = L.m_first;
    while (p != 0) {
        COUT << p ->m_val;
        if (! (p->next == 0) ) { // ltimo node de la lista?
           COUT << ",";
        }
        p = p->next;
    }
    COUT << ")";
#endif

#if 0
    list::iterator p;     // recorre la lista
    list::iterator endL;  // despus de la lista

    COUT << "(";

    p    = L.begin();  // p.operator= ( Lista.begin() );
    endL = L.end();    // generalmente NULL==0

    for ( ; p != endL; ++p) {
        COUT << ( *p );
        if (p != last(L) ) {
            COUT << ",";
        }
    }

    COUT << ")";
#endif

#if 0
    COUT << "(";
    list::node *p = L.m_first;
    while (p != 0) {
        COUT << p ->m_val;
        if (! (p->next == 0) ) { // ltimo node de la lista?
           COUT << ",";
        }
        p = p->next;
    }
    COUT << ")";
#endif

#if 0
    if ( L.empty() ) {
        COUT << "()";
    }
    else {
        list LO;
        LO = L; // Hace una copia
        COUT << "(";
        while (! LO.empty()) {
            value_type val = LO.front();
            COUT << "  " << val;
            LO.pop_front();
        }
        COUT << ")";
    }
#endif

#if 0
    COUT << "(";
    for (list LO = L; ! LO.empty(); LO.pop_front() ) {
        value_type val = LO.front();
        COUT << "  " << val;
    }
    COUT << ")";
#endif

    return COUT;
}  // operator <<

#include <assert.h>

/// Lee del flujo de texto \c "CIN" el valor de \c "L".
/// - Se presume que el valor fue grabado con \c operator<<().
///   - "()"
///   - "(1 2 3)"
///   - "(1 2 3 77 888 999)"
std::istream& operator >> (std::istream &CIN, list& L) {
    L.clear(); // borra todo el valor anterior

    char ch;
    do {
        ch = CIN.get(); // lee el '(' de la pura izquierda
    } while ( ch != '(' );
    assert( '(' == ch );
    list::value_type v; // temporal para leer cada valor de la lista

    for (;;) {
        ch = CIN.peek(); // lee el delimitador o el ')'
        if (ch == ')') {
            ch = CIN.get(); // lee el ')' del final
            break;
        }
        else {
            CIN >> v;
            L.push_back( v );
        }
    }
    return CIN;
}  // operator >>

} // namespace ADH  // } // namespace ADH

// EOF: ADH_list.cpp
