/* ADH_map.cpp (C) 2002  adolfo@di-mare.com  */

#include "ADH_map.h"

namespace ADH { // namespace ADH {

/** Verifica la invariante del diccionario.
    - Regresa \c "true" si el diccionario \c "DDD" contiene un valor correcto
    - Podra retornar \c "true" para un diccionario lista cuyo valor est corrupto
    - Podra no retornar si el diccionario tiene su valor corrupto

    - Como los valores del diccionario estn ordenados, verifica
      que la lista que DDD contiene est ordenada

    \par Rep Diagrama de la clase
    \code
    +-------+                                    NULL
    | +---+ |    +-----+---+   +-----+---+   +-----+---+
    | | *-|-+--->| v_1 | *-|-->| v_2 | *-|-->| v_3 | 0 |
    | +---+ |    +-----+---+   +-----+---+   +-----+---+
    +-------*     m_val next
    \endcode

    - El diccionario est implementado como un envoltorio alrededor de la lista.
    - Por eso es muy lenta la principal operacin de acceso del diccionario, DDD[].
    - Una implementacin ms eficiente usara un rbol balanceado para disminuir
      el tiempo de bsqueda e insercin.

    \pre Obliga al programador cliente a implementar la funcin
          \c bool check_ok( const T & );
*/
bool check_ok( const map & DDD ) {
    if ( ! check_ok( DDD.m_L ) ) {
        return false;
        // sobra invocar aqu a check_ok( const DDD:value_type& )
        // porque check_ok( const ADH_list& ) verifica que cada uno
        // de sus elementos sea correcto.
    }

    // Verifica que los valores de la lista estn ordenados
    list::iterator p;     // recorre la lista m_L
    list::iterator pNext; // ++p == pNext
    list::iterator endL;  // final del diccionario
    p    = DDD.m_L.begin();
    endL = DDD.m_L.end();

    if (p == endL) { // lista vaca
        return true;
    }
    pNext = p;
    pNext++;

    for ( ; pNext != endL; ++p, ++pNext ) {
        if ( (*p).first > (*pNext).first ) { // compara llaves
            return false;
        }
    }
    return true;

/*  NOTA
    Esta implementacin no cumple con la invariante de la clase
    porque los valores no se guardan ordenados en la lista "m_L".
    Esto podra remediarse con facilidad, pero en ese caso es
    mejor usar la plantilla <map> de STL, porque lo importante
    en esta implementacin es mostrar la escencia de lo que es
    la clase diccionario, y para hacerlo es menos complicado no
    ordenar los valores dentro de la lista "m_L".
*/
}

/**  Retorna una referencia al valor asociado a \c "k" en el diccionario
     - Si el valor no est en el diccionario, lo agrega.
     - DDD["banana"].
*/
map::mapped_type & map::operator[] (const map::key_type & k) {
    list::iterator p;     // recorre la lista m_L
    list::iterator endL;  // final del diccionario

    p    = m_L.begin();
    endL = m_L.end();

    for ( ; p != endL; ++p ) {
        if ( (*p).first == k ) {
//          return p->second;
//          return (p.operator->()).second;
//          return (p.operator->())->second;
            return (*p).second;
        }
    }

    // Como no aparece la llave "k", agrega un "pair" para "k"
    pair tmp; // "tmp" queda inicializado con el constructor de vector

    tmp.first = k;
//  tmp.second = mapped_type();
//  tmp.second ==> fue inicializado por su constructor de vector

    m_L.push_front( tmp ); // agrega "tmp" al principio de m_L
    p = m_L.begin();       // el valor insertado queda de primero

    return  p.operator*().second;
}

/** Encuentra el sitio en donde est el valor asociado a \c "k".
    - Si no existe en el diccionario, retorna \c "end()"
*/
map::iterator map::find(const map::key_type & k) {
    list::iterator p;     // recorre la lista m_L
    list::iterator endL;  // despus del diccionario
    map::iterator  res;
    p    = m_L.begin();
    endL = m_L.end();

    for ( ; p != endL; ++p ) {
        if ( (*p).first == k ) {
//          return p->second;
//          return (p.operator->()).second;
//          return (p.operator->())->second;
            res.m_q = p;
            return res;
        }
    }
    res.m_q = endL;
    return res;
}

/** Agrega la pareja \c "V" al diccionario \c "*this"
    - Si el valor \c "V.first" ya estaba almacenado en el diccionario
      deja inalterado el valor de \c "V.second".
    - Retorna la posicin del valor en el diccionario.
*/
map::iterator map::insert (const value_type & V) {
    list::iterator p;     // recorre la lista m_L
    list::iterator endL;  // despus del diccionario
    map::iterator  res;
    p    = m_L.begin();
    endL = m_L.end();

    for ( ; p != endL; ++p ) {
        if ( (*p).first == V.first ) {
            res.m_q = p; // ignorar duplicados
            return res;
        }
    }

    m_L.push_front( V );   // agrega "V" al principio de m_L
    res.m_q = m_L.begin(); // el valor insertado queda de primero
    return res;

/*  NOTA:
    Esta implmentacin est incorrecta, porque en la invariante
    se define que los valores almacenados en la lista deben
    quedar ordenados, pero insert() no los ordena. Esto quiere
    decir que si se ejectura check_ok() luego de insert(), en
    algunos casos se obtendra "false".
    - Seguro algn estudiante corregir pronto esta carencia.
*/
}

/** Graba el valor de \c "M" en el flujo \c "COUT".
    - En particular, este es el operador que se invoca
      cuando se usa, por ejemplo, este tipo de instruccin:
          cout << M << MO;
*/
std::ostream& operator<< (std::ostream &COUT, const map& M) {
    COUT << "(";
    list::iterator p;     // recorre la lista m_L
    list::iterator endL;  // final del diccionario

    p    = M.m_L.begin();
    endL = M.m_L.end();

    for ( ; p != endL; ++p ) {
        COUT << (*p);
        list::iterator q = p; ++q;
        if ( q != endL ) {
            COUT << ',';
        }
    }

    COUT << ")";

    return COUT;
}  // operator <<

} // namespace ADH  // } // namespace ADH

// EOF: ADH_map.cpp
