// Astring.cpp        (c) 2005 adolfo@di-mare.com

/** \file  Astring.cpp
    \brief Implementaciones para la clase \c "Astring"

    \author Adolfo Di Mare <adolfo@di-mare.com>
    \date   2005

    - Why English names??? ==> http://www.di-mare.com/adolfo/binder/c01.htm#sc04
*/
#include <assert.h> // NDEBUG && assert()
#include "Astring.h"

char Astring::null = 0;

/** Verifica la invariante de la clase \c Astring.
   - <code> this->m_s == 0 </code>==> la hilera est vaca
   - <code> this->m_s != 0 </code>==> *(m_s) es un vector en
     memoria dinmica de n+1 chars, con <code> n == strlen(m_s) </code>
   - <code> Astring::null == 0 </code> siempre
   - <code> Astring::null </code> es la hilera nula \c ""

\par Dibujo del Rep de la clase \c Astring
\code
      s
    +---+
    | *-|-------------------------->"BananaRoja"
    +---+

- Cada instancia de una hilera contiene, nicamente, un puntero
  a la hilera, la que siempre est en memoria dinmica.
\endcode
*/
bool check_ok( const Astring& STR ) {
    if (! (Astring::null == 0) ) {
        return false; /// INVARIANTE: <code> Astring::null == 0 </code>
    }
    else if ( ! (&STR != 0) ) {
        return false; /// - Invariante: el objeto no puede estar almacenado en una posicin nula.
    }
    else if (STR.m_s == 0) {
        return true;
    }
    else if (! (STR.m_s[0] != 0) ) { // la hilera nula NO se representa as
        return false;  /// INVARIANTE: <code> STR.m_s[0] != 0 </code>
    }

    for (const char* p = STR.m_s; ( *p != 0 ); ++p) { }
    return true;
} // check_ok()

/// Cambia el valor de \c "*this" para que sea igual a \c "str"
Astring& Astring::operator = (const char *str) {
    if ( str == m_s ) {
        return *this; // evita auto-copia o copia doble
    }
#undef USE_LA_VIEJA
#ifdef USE_LA_VIEJA

    if (m_s != 0) {  // borra el valor anterior
        delete [] m_s;
        m_s = 0;
    }
    assert( m_s == 0 );

    if (str != 0) {          // no est vaca y...
        if (str[0] != 0) {   // no es la hilera nula
            m_s = new char [1 + strlen(str)];
            strcpy( m_s, str ); // sera ms eficiente usar memcpy()
        }
    }

#else
    // Esta implementacin es "ms inteligente":
    // - si las 2 hileras tienen el mismo largo, slo copia sin borrar

    if (m_s == 0) {      // sobra borrar el valor anterior
        if (str != 0) {
            if (str[0] != 0) { // no es la hilera nula
                size_t str_len = strlen(str);
                m_s = new char  [1 + str_len] ; // memcpy() generalmente es
                memcpy(m_s, str, 1 + str_len) ; // ms eficiente que strcpy()
            }
        }
    }
    else { // s != ""
        if (str == 0) {
            delete [] m_s;       // slo  borra
            m_s = 0;
        } else if (str[0] == 0) {
            delete [] m_s;       // slo  borra
            m_s = 0;
        } else {
            size_t s_len   = strlen(m_s);
            size_t str_len = strlen(str);
            if (str_len != s_len) { // s hace falta borrar
                delete [] m_s;       // borra y copia
                m_s = new char [1 + str_len];
            }
            memcpy(m_s, str, 1 + str_len); // copia
        }
    }
#endif
    return *this;
}

/// Agrega \c "str" al final de \c "*this"
Astring& Astring::operator += (const char * str) {
    if (str == 0) {
        return *this;  // no hay nada que agregar
    }
    if (str[0] == 0) {
        return *this;
    }

    if (m_s == 0) {  // copia directo la hilera
        size_t str_len = strlen(str);
        m_s = new char [1 + str_len];
        memcpy( m_s, str, 1+str_len );
//      m_s[str_len] = 0; // no hace falta
    }
    else {
        size_t s_len   = strlen(m_s);
        size_t str_len = strlen(str);

        char * tmp = new char [s_len + str_len + 1];

        memcpy(tmp,       m_s,  s_len);
        memcpy(tmp+s_len, str, str_len+1);
//      tmp[s_len + str_len] = 0; // menos eficiente

        delete [] m_s;
        m_s = tmp;
    }
    return *this;
}

/** Cambia \c "*this" para que tenga una copia de la subhilera de \c "str" de
    \c "n" caracteres comenzando en \c "pos"
    - Si no hay \c "n" caracteres a partir de \c "pos" en \c "str",
      copia slo los primeros.
    - Si \c "str" tiene cero o menos de \c "pos" caracteres,
      en \c "*this" queda la hilera nula.
    - Funciona an si \c "&str" es \c "this".
*/
Astring& Astring::assign(const Astring& str, unsigned pos, unsigned n) {
    unsigned str_len = str.size();
    if (str_len <= pos) { // cubre el caso (str_len == 0)
        if (m_s != 0) {
            delete [] m_s;
            m_s = 0;
        }
        return *this;
    }

    if (n > str_len-pos) { // <==> (n+pos > str_len)
        n = str_len-pos;
    }

    char * tmp = new char [1 + n];
    memcpy(tmp, str.m_s+pos, n);
    tmp[n] = 0; // EOS

    if (m_s != 0) {
        delete [] m_s;
    }
    m_s = tmp;
    return *this;
}

/** Asigna a \c "*this" la hilera formada por los primeros \c "n"
    caracteres de \c "str".
    - Siempre copia \c "n" caracteres an si <code> strlen(str) < n </code> .
*/
void Astring::Set(const char *str, size_t n) {
    *this = "";
    if (n>0) {
        m_s = new char [1 + n];
        memcpy(m_s, str, n);
        m_s[n] = 0;
    }
}

/// Especifica un nuevo tamao para la hilera, eliminando o
/// o agregando elementos como sea requerido.
/// - Agrega el caracter \c "ch" al final si hace falta.
void Astring::resize( size_t n, char ch ) {
    char* tmp = m_s; // lo borra despus de usarlo
    size_t len;
    if (m_s == 0) {
        len = 0;
    } else {
        len = strlen(m_s);
    }

    if (n == len) {
        return; // ya tiene el tamao correcto
    } else if ( n == 0 ) {
       delete[] m_s; // assert( m_s != 0 );
       m_s = 0;
        return;
    }

    m_s = new char [1 + n]; // "tmp" conserva el valor original

    if (len < n) {
        memcpy( m_s,     tmp,   len );
        memset( m_s+len, len-n, ch  );
    } else {
        memcpy( m_s, tmp, n ); // assert( len > n );
    }
    m_s[n] = 0;

    if ( tmp != 0 ) {
        delete [] tmp;
    }
}

#include <assert.h>

/** Extrae de "CIN" un rengln completo y lo pone en la hilera \c "res"

    - Se detiene cuando llega al fin de lnea, en cuyo caso elimina el fin de lnea \c '\\n'
    - Se detiene cuando se le acaba \c "CIN"
    - Sirva para usar \c <stream>'s lo menos posible

    \par NOTA: Este extractor funciona de manera DIFERENTE al de la biblioteca
*/
std::istream& operator >> ( std::istream& CIN, Astring& res ) {
    const unsigned INPUT_STRING_SIZE = 128; // Tamao mximo del bloque de hilera que jala Get_Line()
    char  buffer  [INPUT_STRING_SIZE]; // ac se leen los bloques de letras
    res = "";

    while ( CIN.good() ) { // aqu jala un bloque de letras
        // istream& get(char*, int len, char = '\n')
        CIN.get(buffer, INPUT_STRING_SIZE, '\n');
        unsigned n = CIN.gcount();
        if ( n == 0 ) { // la lnea est vaca
            char ch;
            CIN.get(ch); // se brinca el fin de lnea
            assert( ch == '\n' );
            return CIN;
        } else {
            res += buffer; // agrega en bloquetones lo leido
            if (n != INPUT_STRING_SIZE-1) {
                char ch;
                CIN.get(ch); // se brinca el fin de lnea
                assert( ch == '\n' );
                return CIN;
            }
        }
    }
    return CIN;
} // operator >> ()

// EOF: Astring.cpp
